# PSA1, Python, vezbe


# Python, aliases (aliasi), identiteti objekata
# Podjimo od objekta a:
a = 3
type(a)
# Uvodimo i objekat b:
b = a
# operator 'is' uporedjuje identitete objekata, u ovom slucaju objekata a i b; 
# rezultat Bulova promenljiva (True/False)
a is b
# Rezultat je ovde True, sto znaci da su ovo dva ista objekta samo su drugacija
# simbolicka imena (aliasi) koja su im dodeljena; u pitanju su aliasi (a i b) 
# koji ukazuju na isti objekat u memoriji.

# id(a) - pokazuje identitet objekta a; Svaki objekat ima identitet, tip i 
# vrednost (tip i vrednost se vide u Variable explorer). Identitet je 
# jedinstven i konstantan broj (int) za vreme postojanja objekta a; Moze se 
# shvatiti i kao adresa objekta u memoriji.
id(a)
# id(b) - pokazuje identitet objekta b
id(b)
# Ako su identiteti objekata isti, sto je ovde slucaj, mi znamo da je to isti 
# objekat, samo ima drugo simbolicko ime, drugi alias (jedan je a, drugi je b)

# operator == uporedjuje vrednosti objekata:
a == b
 
# Uvecajmo sada b za 1:
b += 1
print(b)
print(a)
# da li je sada a==b?
a == b
# a da li je identitet a jednak identitetu b?
a is b
id(a)
id(b)
# id(b) se promenio, id(a) ostao isti. Objekat b je prebacen na neku drugu 
# lokaciju i on vise nije isto sto i a.



# Python, aliases sa listama
# Probajmo isto samo sada na primeru listi a ne celih brojeva (int);
# Podjimo od objekta tipa liste a:
a = [1, 2, 3]
print(a)
# Uvedimo i objekat b:
b = a
print(b)

# Koji su im identiteti:
id(a)
id(b)
# Dakle, identitet a je i ovde jednak identitetu b
a is b
# a i vrednosti su ime jednake:
a == b

# Promenimo sada element liste b (kao i malopre na primeru int):
b[1] = 0
print(b)

# Paznja! Sta je sa listom a?
print(a)
# Ali listu a nismo menjali, vec samo listu b?! U primeru sa int a nije 
# menjao vrednost proomenom b a u primeru sa listom jeste!
# Sta kazu identiteti?
id(a)
id(b)

# da li je vrednost a jednaka vrednosti b?
a == b
# da li je identitet a jednaka identitetu b?
a is b
# Zakljucak: Python razlicito postupa sa imenima objekata (aliasima) kada su u 
# pitanju liste i kada su u pitanju int!

# Vratimo se na pocetno a, i posmatrajmo novi primer sa identitetima:
a = [1, 2, 3]
b = a
c = a[:]

# poredimo vrednosti a i b, a i c, b i c, respektivno:
a == b
a == c
b == c

# a sta je sa identitetima?
a is b
a is c
# c je samo kopija element po element cele liste a!
b is c
id(a)
id(b)
id(c)
# Ako kopiramo listu i nameravamo da je menjamo a necemo da se originalna 
# lista menja tada je kopiramo sa dvotackom [:]; identiteti te liste i one
# originalne nisu isti!

# Promenimo sada listu c:
c[0] = 'd'
print(c)
print(a)
print(b)
# sada ni sadrzaji, tj. vrednosti, listi a i c nisu isti:
c == a



# Python, 'matrice'
# Matrice u Pythonu nisu bas podrzane na nacin kako smo to matematicki
# navikli, ali zato postoji modul (NumPy) koji to bas tako podrzava;
# Bez modula, matrice se definisu kao liste listi, npr.
a = [[1, 2], [3, 4]]
print(a)
# vidimo da i nema bas oblik matrice...
len(a)
# len(a) je broj listi unutar liste a; da li su to kolone ili vrste to je na 
# nama da interpretiramo kako hocemo...

# Kako pristupamo elementima matrice (a one su liste listi)?
# Broj elemenata u podlisti a[1]:
len(a[1])
# podlista a[1]
print(a[1])
# element indeksa 1 u podlisti 1
print(a[1][1])
# element indeksa 0 u podlisti 0
print(a[0][0])

# veca matrica 
a = [[1, 2, 5], [3, 4, 6]]
print(a)
len(a)
len(a[0])

# Liste nisu optimalno resenje za rad sa matricama jer matrice sadrze objekte 
# istog tipa za razliku od listi; ovo se posebno odnosi na operacije nad 
# matricama.
# Kako se pristupa elementima matrice? Npr. kako pristupiti elementu 5 u 
# matrici a = [[1, 2, 5], [3, 4, 6]]? On je u okviru nultog elementa liste a
# a u okviru njega je treci element podliste, sto znaci indeksa 2, pa je:
print(a[0][2])
# dakle, indeksira se objekat, pa unutar tog objekta podobjekat, ...
# do 6 se npr. dolazi:
print(a[1][2])
# do 3 se dolazi:
print(a[1][0])



# Python, inicijalizacija 'nizova' (preko listi)
# Inicijalizacija prazne liste je:
a = []
print(a)

# Overloadovan operator *; ovo je operator mnozenja ali za razlicite objekte
# moze da se definise na razlicite nacine - overloaduje; lista sa 10 
# nultih clanova:
a = [0]*10
print(a)
type([0])

# primer: da li je ovo tuple sa 10 clanova?
a = (0)*10
print(a)
type((0))
# nije, vec je int!

# Ovo je tuple sa 10 clanova:
a = (0,)*10
print(a)
type((0,))


# Jos jedan primer - Inicijalizacija 3x3 matrice sa jedinicama:
a = [[1]*3]*3
print(a)




# Python, for petlja
# Kao i u drugim programskim jezicima, koristi se za ponavljanje dela koda
# kada je unapred poznat broj izvrsavanja (ponavljanja); cesto se definise 
# preko funkcije range()
# Primer: iteriranje po sekvenci koju formira funkcija range(); primer 
# ispisuje brojeve u obliku "tekuci broj sekvence + 1 / ukupna duzina sekvence": 
a = range(10)
for i in a:
    print(i + 1, '/', len(a))

# Umesto a = range(10) radimo iteriranje preko liste a = list(range(10));
# Dobija se isto a potrosili vise memorije jer smo 'a' pretvorili u listu; 
# Dakle, treba raditi kao gornji primer, tj. a = range(10) 
a = list(range(10))
for i in a:
    print(i + 1, '/', len(a))



# Python, for petlja, iteriranje po stringu (over string)
# Malo neobicno ali for petlja moze da se primeni po svemu sto je iterabilno
# tj. ima elemente po redosledu do nekog konacnog broja; primer: iteriranje
# po stringu a:
a = 'softverski alati'
for znak in a:
    print(znak)



# Python, if-else struktura
# Primer: u reci 'abrakadabra' zameniti sva slova 'a' donjom crtom _ 
a = 'abrakadabra'
b = ''
for znak in a:
    if znak != 'a':
        b += znak
    else:
        b += '_'
print(b)



# Python, varijanta if strukture: if-elif-else
# prosirenje za ilustraciju if-elif-else: u reci 'abrakadabra' zameniti sva 
# slova 'a' donjom crtom _ a sva slova 'k' zvezdcom *
a = 'abrakadabra'
b = ''
for znak in a:
    if znak == 'a':
        b += '_'
    elif znak == 'k':
        b += '*'
    else:
        b += znak
print(b)



# Python, for petlja sa continue 
# Primer ilustruje primenu komande 'continue' u for petlji; u listi koja
# sadrzi brojeve od 0 do 99 proverava se da li je broj deljiv sa 4 (ostatak
# pri deljenju razlicit od 0) pa ako jeste izbacuje ga iz liste;
a = []
for i in range(100):
    if i % 4 != 0:
        a.append(i)
        continue
    print('izbacen', i)
print()
print(a)

# dakle, komanda 'continue' nakon dodavanja broja nedeljivog sa 4 u listu
# ne procesira dalje telo petlje vec ide na novo 'i'


# Isto samo realizovano preko if-else (probati i kao skript, pokrenuti preko 
# konzole program - run izbacivanje4)
a = []
for i in range(100):
    if i % 4 != 0:
        a.append(i)
    else:
        print('izbacen', i)
print()
print(a)


# Primena 'break' - prekid petlje
# Primer za break: u primeru sa izbacivanjem broja 4 iz liste se ne ide do 
# kraja liste vec kada broj postane deljiv sa 77 program stane:
a = []
for i in range(100):
    if i % 4 != 0:
        a.append(i)
        continue
    print('izbacen', i)
    if (i+1) % 77 == 0:
        break
print()
print(a)



# Python, I/O, input 
# Ako vrednosti svih promenljivih nisu definisane u samom programskom kodu već
# je potrebno da neke unese sam korisnik, koristi se funkcija input():
x = input('unesi neki podatak: ')
print(type(x))
print(x)

x = int(input('unesi ceo broj: '))
print(x, type(x))
# Dakle, svaki input koji mozemo uneti je tipa string, a sve ostalo 
# moramo sami da konvertujemo, kao sto je u primeru konvertovan u int 



# Python, I/O, print i formatted print;
# Za prikaz rezultata na standardni izlaz (ekran), koristi se funkcija print():
print('Praktikum softverski alati 1')
print('{')
print('{}')

# Formatirani print - za slucaj kada mi zelimo da formatiramo print kako mi
# hocemo. Oblika je:
# print(str.format())
# na mestu viticastih zagrada upise ono unutar format()
print('{}'.format('Pera'))
print('U toku je cas iz predmeta {}' .format('Praktikum softverski alati 1'))

# formatirani string moze i preko kljucne reci, ovde je to rec predmet:
print('Nadam se da {predmet} nije dosadan predmet' .format(predmet = "PSA1"))

# Formatirani print sa vise pojmova:
print('{}, {}'.format('Mika', 'Laza'))
# a moze i sa redosledom ispisa pojmova
print('{0}, {1}'.format('Mika', 'Laza'))
print('{1}, {0}'.format('Mika', 'Laza'))
print('{1}, {1}'.format('Mika', 'Laza'))
print('{1}, {1}, {0}, {1}'.format('Mika', 'Laza'))



# Python, formatted print, brojevi
print('p = {}, q = {}'.format(3, 4))
print('p = {}, q = {}'.format(3.0, 4))
print('p = {1}, q = {0}'.format(3, 4))
print('p = {}, q = {}'.format(3.0000, 4))
print('p = {}, q = {}'.format(3., 4))
print('p = {}, q = {}'.format(float(3), 4))
# Dakle, u primerima iznad zapis brojeva je po Python standardu.
# Ali ako mi hocemo tacno da specificiramo formate zapisa brojeva (primer sa
# brojem pi):
import math
print('pi = {:.2f}'.format(math.pi))
# : je separator koji kaze da se posle ove dvotacke nalazi specifikacija 
# formata; ovde je to float (slovo f) u zapisu sa 2 decimale; ovaj format je 
# delimicno specificiran jer se ne kaze koliko ce mesta da zauzme broj vec 
# samo da ce posle decimalne tacke imati dva decimalna mesta.

# isti primer samo hocemo pi u 5 decimala:
print('pi = {:.5f}'.format(math.pi))

# a moze i kompletna specifikacija tj. sirina polja zapisa: 
# primer za ovo: ukupan prostor je 10 mesta od cega je 5 mesta iza decimalne 
# tacke
print('pi = {:10.5f}'.format(math.pi))

# formatiranje celih brojeva: separator (dvotacka) pa broj mesta za broj, i
# specifikacija da je ceo broj, slovo d:
print('n = {:10d}'.format(100))
print('n = {:4d}'.format(100))
print('n = {:3d}'.format(100))
print('n = {}'.format(100))

# A ako smo pogresili? 1 cifra za broj 100?! - ignorise nas:
print('n = {:1d}'.format(100))



# Python, stringovi i operacije nad stringovima;
# Stringovi predstavljaju niz simbola (karaktera) koji se predstavljaju
# između apostrofa:
a = 'Petar'
b = 'Marko'

# len() - broj karaktera u stringu
len(a)
len(b)

# + konkatenacija
a + b
b + a
# ocigledno ne vazi komutativnost jer je ovo nadovezivanje a ne sabiranje

# * ponavljanje
a * 3
3 * a

# moguce praviti i izraze
(a + 2 * b)  * 3

# oduzimanje stringova ne postoji (nije definisano)
#a - b 

# operator poredjenja stringova
# == jednako
a == a
a == b

# indeksiranje elementa (isto ide on nule)
a[2]

# izdvajanje opsega, tj. podstringa (kao i kod listi): a[i:j]: od i do j-1
a[1:4]

# in - nalazi li se element u stringu? 
't' in a
'm' in a

# < manje
a < b
b < a
'1' < '2'


# Python, Modul string
import string as s
# Sta se nalazi u tom modulu - preko dir():
dir(s)
# Pomoc preko help():
help(s)
# sva ascii slova se dobijaju preko
s.ascii_letters
# svi karakteri iz ascii koji su printabilni
s.printable

# sva velika slova
s.ascii_uppercase

# sve cifre
s.digits

# sve oktalne cifre
s.octdigits

# brisemo modul
del s


# Python, metode nad stringovima (metode su jer imaju zagrade)
# velika slova
'Pera'.upper()

# mala slova
'Pera'.lower()

# centiranje na polje koje ima 20 karaktera (ostatak prazno)
'Pera'.center(20)

# isalpha() da li se string sastoji samo iz slovnih karaktera?
'Pera'.isalpha()
'Pera6'.isalpha()

# veliko pocetno slovo (prvi simbol celog stringa)
'praktikum softverski alati'.capitalize()

# velika pocetna slova svakog podstringa (separator je blanko)
'praktikum softverski alati'.title()

# lista podstringova unutar stringa (separator je blanko)
'Praktikum softverski alati'.split()
# ili po nekom drugom separatoru...
'Praktikum softverski alati'.split(',')
'Praktikum,softverski,alati'.split(',')
'Praktikum, softverski, alati'.split(',')



# Python, datoteke (files)
# Kada se radi sa fajlovima, fajl prvo mora da se otvori i na kraju da se 
# zatvori a izmedju vec nesto radimo sa tim fajlom;
# Otvaranje datoteke:
f = open('proba.txt', 'w')
print(f)
# dakle, fajl se zove proba.txt, mod write 'w' je za zapis (posto nema 
# specifikaciju zapis je txt a ne binarni), i na kraju je encoding; 
# encoding zavisi od operativnog sistema i za Windows je cp1252
type(f)
f
# Upis u datoteku:
f.write('prvi red\n')
# \n oznacava novu liniju; 
# program je vratio je broj 9 a to je broj karaktera koji je upisan

f.write('drugi red\n')
f.write('treci red\n')

# f.close - zatvaranje je metod a ne parametar pa moraju zagrade ()
f.close()
type(f)
# ostalo isto kao na pocetku 
f

# Citanje datoteke
# Sada da otvorimo taj fajl proba.txt ali u read 'r' modu:
f = open('proba.txt', 'r')
f
# kao sto se vidi u zapisu sada je mod read

# ako hocemo da citamo sve (ceo fajl je pretvoren u jedan string):
f.read()
f.close()

# citanje sa specificiranim brojem karaktera
f = open('proba.txt', 'r')
f.read(9)
f.read(10)
f.read(10)
f.close()

# citanje sa specificiranim brojem karaktera (premasenim)
f = open('proba.txt', 'r')
f.read(11)
f.close()

# Citanje preko readline
f = open('proba.txt', 'r')
f.readline()
# dakle, cita do oznake za new line (\n) pri cemu se i taj \n uracunava u taj 
# prvi string pa do sledeceg \n
f.readline()
# pa do sledeceg \n
f.readline()
f.close()


# Datoteka kao iterabilni objekat
f = open('proba.txt', 'r')
for line in f:
    print(line)
f.close()
# Dakle, ovde je ceo fajl tretiran kao objekat f koji ima clanove koji su 
# iterabilni: clanovi su linije tj. deo fajla do \n (prva, druga, treca); 
# radio je readline-readline-readline;
    


# Python, map
# Zelimo da neki niz brojeva isprocesiramo nekom funkcijom; to se svakako 
# moze uraditi preko petlje, ali u Pythonu za ovo postoji vec gotova funkcija
# i ona se zove 'map'.
# Uzmimo kao primer niz brojeva i njihove kvadrate - pravimo funkciju sq(): 
x = range(11)
print(x)
def sq(a):
    return a**2
   
# map(a, b) - prvi argument je nasa funkcija sq a drugi promenljiva x
y = map(sq, x)
list(y)

# lista u jednom potezu:
y = list(map(sq, x))
print(y)

# ako hocemo tuple a ne listu:
y = tuple(map(sq, x))
print(y)

# pomoc za objekt tipa liste
help([])

# pomoc za objekt tipa tuple
help(())

# pomoc za objekt tipa dictionary 
help({})



# Python, recnik (dictionary)
# Dictionary - asocijativni nizovi: indeks je u vidu kljucne reci koja bi 
# trebala da nas asocira na vrednost;
# Dictionary je poseban tip objekta, tip 'dict'; izmedju viticastih zagrada 
# nalaze se zapisi koji sadrze dva dela: kljuc tj. key (koji je string)
# i vrednosti tj. value, u formatu key:value;
# Primer telefonskog imenika vasih profesora sa PSA1 kao tipa dictionary:
t = {'mira': 400, 'jelena': 348, 'milos': 361}
type(t)
print(t)

# Ovaj recnik (imenik) ima sledece kljuceve (tj. adrese) ciji su telefoni:
t.keys()

# Ovaj recnik (imenik) ima sledece vrednosti tj. same brojeve telefona:
t.values()
# i keys i values su orderovane liste

# Primena: koji je telefon od prof. Milosa?
t['milos']
# ili od prof. Jelene?
t['jelena']
# a ja cu da promenim broj:
t['mira'] = 000
print(t)
t['mira'] = '000'
print(t)
# Kao i kod lista, i kod dictionery objekti mogu biti razlicitog tipa: value  
# kod mire bio int 0 a sad je string '000'; ostali bili i ostali int
type(t['mira'])
type(t['milos'])
type(t['jelena'])

# Sad cemo da izbrisemo prof. Jelenu iz imenika:
del t['jelena']
print(t)

# Moze i da se pita da li je neko u telefonskom imeniku:
'mira' in t
# Ako stavimo samo veliko pocetno slovo to tretira kao drugaciji key - case 
# sensitive
'Mira' in t
'jelena' in t 



# Python, sets
# Set kao poseban tip objekta; takodje viticaste zagrade kao i kod dictionery 
# ali...
a = 'abracadabra'
b = 'alacazam'
x = set(a)
print(x)
# Dakle, set() nam iz nekog brebrojivog iterabilnog skupa izdvaja razlicite 
# elemente
y = set(b)
print(y)
type(y)

# Operacije nad setovima; posto je ovo novi tip objekta, onda su operacije
# nad njim drugacije definisane;
# zbir dva skupa (ne postoji ovaj operator za setove):
# x + y

# razlika dva skupa - iz x oduzmemo elemente koji su u y: (u primeru: c i a)
x - y
y - x

# Unija dva skupa
x | y

# Presek dva skupa
x & y

# XOR - oni clanovi unije koji preostaju kada se izdvoje zajednicki
x ^ y





